package nxt.addons;

import nxt.Nxt;
import nxt.NxtException;
import nxt.blockchain.Block;
import nxt.blockchain.Chain;
import nxt.blockchain.ChainTransactionId;
import nxt.blockchain.Transaction;
import nxt.http.API;
import nxt.http.APIServlet;
import nxt.http.APITag;
import nxt.http.JSONData;
import nxt.http.JSONResponses;
import nxt.http.ParameterParser;
import nxt.lightcontracts.ContractReference;
import nxt.util.Convert;
import org.json.simple.JSONObject;
import org.json.simple.JSONStreamAware;

import javax.servlet.http.HttpServletRequest;

import static nxt.addons.ContractRunner.INVOCATION_TYPE.REQUEST;
import static nxt.addons.ContractRunner.INVOCATION_TYPE.VOUCHER;

class ContractRunnerAPIs {

    public static class TriggerContractByTransactionAPI extends APIServlet.APIRequestHandler {
        private final ContractRunner contractRunner;

        TriggerContractByTransactionAPI(ContractRunner contractRunner, APITag[] apiTags, String... origParameters) {
            super(apiTags, origParameters);
            this.contractRunner = contractRunner;
        }

        @Override
        protected JSONStreamAware processRequest(HttpServletRequest req) throws NxtException {
            API.verifyPassword(req);
            byte[] triggerFullHash = ParameterParser.getBytes(req, "triggerFullHash", true);
            Chain chain = ParameterParser.getChain(req);
            boolean isApply = "true".equalsIgnoreCase(req.getParameter("apply"));
            boolean isValidator = "true".equalsIgnoreCase(req.getParameter("validate"));
            Transaction transaction = Nxt.getBlockchain().getTransaction(chain, triggerFullHash);
            if (transaction != null) {
                return contractRunner.processTransaction(transaction, isApply, isValidator);
            }
            return contractRunner.generateErrorResponse(1001, "Unknown transaction %d:%s", chain.getId(), Convert.toHexString(triggerFullHash));
        }
    }

    public static class TriggerContractByHeightAPI extends APIServlet.APIRequestHandler {
        private final ContractRunner contractRunner;

        TriggerContractByHeightAPI(ContractRunner contractRunner, APITag[] apiTags, String... origParameters) {
            super(apiTags, origParameters);
            this.contractRunner = contractRunner;
        }

        @Override
        protected JSONStreamAware processRequest(HttpServletRequest req) throws NxtException {
            API.verifyPassword(req);
            int height = ParameterParser.getHeight(req);
            String contractName = req.getParameter("contractName");
            boolean isApply = "true".equalsIgnoreCase(req.getParameter("apply"));
            Block block = Nxt.getBlockchain().getBlockAtHeight(height);
            return contractRunner.processBlockContract(block, contractName, isApply, false, null);
        }

        @Override
        protected boolean isChainSpecific() {
            return false;
        }
    }

    public static class TriggerContractByRequestAPI extends APIServlet.APIRequestHandler {
        private final ContractRunner contractRunner;

        TriggerContractByRequestAPI(ContractRunner contractRunner, APITag[] apiTags, String... origParameters) {
            super(apiTags, origParameters);
            this.contractRunner = contractRunner;
        }

        @Override
        protected JSONStreamAware processRequest(HttpServletRequest req) throws NxtException {
            API.verifyPassword(req);
            ContractRunnerConfig config = contractRunner.getConfig();
            if (config instanceof NullContractRunnerConfig) {
                return runnerNotInitializedResponse(config.getStatus()).toJSONObject();
            }
            String contractName = req.getParameter("contractName");
            ContractReference contractReference = ContractReference.getContractReference(config.getAccountId(), contractName);
            if (contractReference == null) {
                return contractRunner.generateErrorResponse(1002, "Contract %s not found", contractName);
            }
            JO setupParams = ContractLoader.getContractSetupParams(contractReference);
            Contract contract = ContractLoader.loadContract(contractReference.getContractId(), setupParams);
            RequestContext context = new RequestContext(req, config, contractName);
            contractRunner.process(contract, context, REQUEST);
            JO jo = context.getResponse();
            if (jo == null) {
                return contractRunner.generateErrorResponse(1002, "Contract %s class %s returned no response", contractName, contract.getClass().getCanonicalName());
            }
            return jo.toJSONObject();
        }

        @Override
        protected boolean isChainSpecific() {
            return false;
        }
    }

    public static final class TriggerContractByVoucherAPI extends APIServlet.APIRequestHandler {

        private final ContractRunner contractRunner;

        TriggerContractByVoucherAPI(ContractRunner contractRunner, String fileParameter, APITag[] apiTags, String... origParameters) {
            super(fileParameter, apiTags, origParameters);
            this.contractRunner = contractRunner;
        }

        @Override
        protected JSONStreamAware processRequest(HttpServletRequest req) throws NxtException {
            API.verifyPassword(req);
            ContractRunnerConfig config = contractRunner.getConfig();
            if (config instanceof NullContractRunnerConfig) {
                return runnerNotInitializedResponse(config.getStatus()).toJSONObject();
            }
            ParameterParser.FileData fileData = ParameterParser.getFileData(req, "voucher", true);
            if (fileData == null) {
                return JSONResponses.INCORRECT_FILE;
            }
            byte[] data = fileData.getData();
            String contractName = req.getParameter("contractName");
            JSONObject voucher = ParameterParser.parseVoucher(data);
            Contract contract = contractRunner.getSupportedContracts().get(contractName);
            VoucherContext context = new VoucherContext(new JO(voucher), config, contractName);
            contractRunner.process(contract, context, VOUCHER);
            JO jo = context.getResponse();
            if (jo == null) {
                return contractRunner.generateErrorResponse(1003, "Contract %s with class %s invoked by account %s returned no response",
                        contractName, contract.getClass().getCanonicalName(), config.getAccountRs());
            }
            if (jo.isExist("transactions")) {
                contractRunner.submitContractTransactions(jo.getArray("transactions").toJSONArray());
            }
            return jo.toJSONObject();
        }

        @Override
        protected boolean isChainSpecific() {
            return false;
        }
    }

    public static class GetSupportedContractsAPI extends APIServlet.APIRequestHandler {
        private final ContractRunner contractRunner;

        GetSupportedContractsAPI(ContractRunner contractRunner, APITag[] apiTags, String... origParameters) {
            super(apiTags, origParameters);
            this.contractRunner = contractRunner;
        }

        @Override
        protected JSONStreamAware processRequest(HttpServletRequest req) throws NxtException {
            API.verifyPassword(req);
            ContractRunnerConfig config = contractRunner.getConfig();
            if (config instanceof NullContractRunnerConfig) {
                return runnerNotInitializedResponse(config.getStatus()).toJSONObject();
            }
            JO response = new JO();
            response.put("contractAccount", config.getAccountRs());
            JA array = new JA();
            for (String name : contractRunner.getSupportedContracts().keySet()) {
                Contract contract = contractRunner.getSupportedContracts().get(name);
                JO contractJson = new JO();
                contractJson.put("name", name);
                contractJson.put("contractClass", contract.getClass().getCanonicalName());
                contractJson.put("params", contract.getContractParams().toJSONObject());
                ContractReference contractReference = contractRunner.getSupportedContractReferences().get(name);
                contractJson.put("contractReference", JSONData.contractReference(contractReference));
                ChainTransactionId contractId = contractReference.getContractId();
                Transaction taggedDataUploadTransaction = contractId.getChildTransaction();
                contractJson.put("uploadTransaction", JSONData.transaction(taggedDataUploadTransaction));
                array.add(contractJson);
            }
            response.put("supportedContracts", array);
            return response.toJSONObject();
        }

        @Override
        protected boolean isChainSpecific() {
            return false;
        }
    }

    public static class ReloadContractRunnerConfigurationAPI extends APIServlet.APIRequestHandler {
        private final ContractRunner contractRunner;

        ReloadContractRunnerConfigurationAPI(ContractRunner contractRunner, APITag[] apiTags, String... origParameters) {
            super(apiTags, origParameters);
            this.contractRunner = contractRunner;
        }

        @Override
        protected JSONStreamAware processRequest(HttpServletRequest req) throws NxtException {
            API.verifyPassword(req);
            contractRunner.loadConfig(req.getParameter("configFileName"));
            JO response = new JO();
            response.put("configLoaded", true);
            return response.toJSONObject();
        }

        @Override
        protected boolean isChainSpecific() {
            return false;
        }
    }

    private static JO runnerNotInitializedResponse(String description) {
        JO response = new JO();
        response.put("errorCode", 1000);
        response.put("errorDescription", description);
        return response;
    }

}
